"""Testing for K-Medoids"""
import warnings
import numpy as np
from unittest import mock
from scipy.sparse import csc_matrix
import pytest

from sklearn.datasets import load_iris, fetch_20newsgroups_vectorized
from sklearn.metrics.pairwise import PAIRWISE_DISTANCE_FUNCTIONS
from sklearn.metrics.pairwise import euclidean_distances, cosine_distances

from numpy.testing import assert_allclose, assert_array_equal

from sklearn_extra.cluster import KMedoids, CLARA
from sklearn.cluster import KMeans
from sklearn.datasets import make_blobs


seed = 0
X = np.random.RandomState(seed).rand(100, 5)

# test kmedoid's results
rng = np.random.RandomState(seed)
X_cc, y_cc = make_blobs(
    n_samples=100,
    centers=np.array([[-1, -1], [1, 1]]),
    random_state=rng,
    shuffle=False,
    cluster_std=0.2,
)


@pytest.mark.parametrize("method", ["alternate", "pam"])
@pytest.mark.parametrize(
    "init", ["random", "heuristic", "build", "k-medoids++"]
)
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_kmedoid_results(method, init, dtype):
    expected = np.hstack([np.zeros(50), np.ones(50)])
    km = KMedoids(n_clusters=2, init=init, method=method, random_state=rng)
    km.fit(X_cc.astype(dtype))
    # This test use data that are not perfectly separable so the
    # accuracy is not 1. Accuracy around 0.85
    assert (np.mean(km.labels_ == expected) > 0.8) or (
        1 - np.mean(km.labels_ == expected) > 0.8
    )
    assert dtype is np.dtype(km.cluster_centers_.dtype).type
    assert dtype is np.dtype(km.transform(X_cc.astype(dtype)).dtype).type


@pytest.mark.parametrize("method", ["alternate", "pam"])
@pytest.mark.parametrize(
    "init", ["random", "heuristic", "build", "k-medoids++"]
)
def test_kmedoid_nclusters(method, init):
    n_clusters = 50

    km = KMedoids(
        n_clusters=n_clusters,
        init=init,
        method=method,
        max_iter=1,
        random_state=rng,
    )
    km.fit(X_cc)
    assert len(np.unique(km.medoid_indices_)) == n_clusters


def test_clara_results():
    expected = np.hstack([np.zeros(50), np.ones(50)])
    km = CLARA(n_clusters=2)
    km.fit(X_cc)
    # This test use data that are not perfectly separable so the
    # accuracy is not 1. Accuracy around 0.85
    assert (np.mean(km.labels_ == expected) > 0.8) or (
        1 - np.mean(km.labels_ == expected) > 0.8
    )


def test_medoids_invalid_method():
    with pytest.raises(ValueError, match="invalid is not supported"):
        KMedoids(n_clusters=1, method="invalid").fit([[0, 1], [1, 1]])


def test_medoids_invalid_init():
    with pytest.raises(ValueError, match="init needs to be one of"):
        KMedoids(n_clusters=1, init="invalid").fit([[0, 1], [1, 1]])


def test_kmedoids_input_validation_and_fit_check():
    rng = np.random.RandomState(seed)
    # Invalid parameters
    msg = "n_clusters should be a nonnegative integer. 0 was given"
    with pytest.raises(ValueError, match=msg):
        KMedoids(n_clusters=0).fit(X)

    msg = "n_clusters should be a nonnegative integer. None was given"
    with pytest.raises(ValueError, match=msg):
        KMedoids(n_clusters=None).fit(X)

    msg = "max_iter should be a nonnegative integer. -1 was given"
    with pytest.raises(ValueError, match=msg):
        KMedoids(n_clusters=1, max_iter=-1).fit(X)

    msg = "max_iter should be a nonnegative integer. None was given"
    with pytest.raises(ValueError, match=msg):
        KMedoids(n_clusters=1, max_iter=None).fit(X)

    msg = (
        r"init needs to be one of the following: "
        r".*random.*heuristic.*k-medoids\+\+"
    )
    with pytest.raises(ValueError, match=msg):
        KMedoids(init=None).fit(X)

    # Trying to fit 3 samples to 8 clusters
    msg = (
        "The number of medoids \(8\) must be less "
        "than the number of samples 5."
    )
    Xsmall = rng.rand(5, 2)
    with pytest.raises(ValueError, match=msg):
        KMedoids(n_clusters=8).fit(Xsmall)


def test_random_deterministic():
    """Random_state should determine 'random' init output."""
    rng = np.random.RandomState(seed)

    X = load_iris()["data"]
    D = euclidean_distances(X)

    medoids = KMedoids(init="random")._initialize_medoids(D, 4, rng)
    assert_array_equal(medoids, [114, 62, 33, 107])


def test_heuristic_deterministic():
    """Result of heuristic init method should not depend on rnadom state."""
    rng1 = np.random.RandomState(1)
    rng2 = np.random.RandomState(2)
    X = load_iris()["data"]
    D = euclidean_distances(X)

    medoids_1 = KMedoids(init="heuristic")._initialize_medoids(D, 10, rng1)

    medoids_2 = KMedoids(init="heuristic")._initialize_medoids(D, 10, rng2)

    assert_array_equal(medoids_1, medoids_2)


def test_update_medoid_idxs_empty_cluster():
    """Label is unchanged for an empty cluster."""
    D = np.zeros((3, 3))
    labels = np.array([0, 0, 0])
    medoid_idxs = np.array([0, 1])
    kmedoids = KMedoids(n_clusters=2)

    # Swallow empty cluster warning
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        kmedoids._update_medoid_idxs_in_place(D, labels, medoid_idxs)

    assert_array_equal(medoid_idxs, [0, 1])


def test_kmedoids_empty_clusters():
    """When a cluster is empty, it should throw a warning."""
    rng = np.random.RandomState(seed)
    X = [[1], [1], [1]]
    kmedoids = KMedoids(n_clusters=2, random_state=rng)
    with pytest.warns(UserWarning, match="Cluster 1 is empty!"):
        kmedoids.fit(X)


@mock.patch.object(KMedoids, "_kpp_init", return_value=object())
def test_kpp_called(_kpp_init_mocked):
    """KMedoids._kpp_init method should be called by _initialize_medoids"""
    D = np.array([[0, 1], [1, 0]])
    n_clusters = 2
    rng = np.random.RandomState(seed)
    kmedoids = KMedoids()
    kmedoids.init = "k-medoids++"
    # set _kpp_init_mocked.return_value to a singleton
    initial_medoids = kmedoids._initialize_medoids(D, n_clusters, rng)

    # assert that _kpp_init was called and its result was returned.
    _kpp_init_mocked.assert_called_once_with(D, n_clusters, rng)
    assert initial_medoids == _kpp_init_mocked.return_value


def test_kmedoids_pp():
    """Initial clusters should be well-separated for k-medoids++"""
    rng = np.random.RandomState(seed)
    kmedoids = KMedoids()
    X = [
        [10, 0],
        [11, 0],
        [0, 10],
        [0, 11],
        [10, 10],
        [11, 10],
        [12, 10],
        [10, 11],
    ]
    D = euclidean_distances(X)

    centers = kmedoids._kpp_init(D, n_clusters=3, random_state_=rng)

    assert len(centers) == 3

    inter_medoid_distances = D[centers][:, centers]
    assert np.all((inter_medoid_distances > 5) | (inter_medoid_distances == 0))


def test_precomputed():
    """Test the 'precomputed' distance metric."""
    rng = np.random.RandomState(seed)
    X_1 = [[1.0, 0.0], [1.1, 0.0], [0.0, 1.0], [0.0, 1.1]]
    D_1 = euclidean_distances(X_1)
    X_2 = [[1.1, 0.0], [0.0, 0.9]]
    D_2 = euclidean_distances(X_2, X_1)

    kmedoids = KMedoids(metric="precomputed", n_clusters=2, random_state=rng)
    kmedoids.fit(D_1)

    assert_allclose(kmedoids.inertia_, 0.2)
    assert_array_equal(kmedoids.medoid_indices_, [2, 0])
    assert_array_equal(kmedoids.labels_, [1, 1, 0, 0])
    assert kmedoids.cluster_centers_ is None

    med_1, med_2 = tuple(kmedoids.medoid_indices_)
    predictions = kmedoids.predict(D_2)
    assert_array_equal(predictions, [med_1 // 2, med_2 // 2])

    transformed = kmedoids.transform(D_2)
    assert_array_equal(transformed, D_2[:, kmedoids.medoid_indices_])


def test_kmedoids_fit_naive():
    n_clusters = 3
    metric = "euclidean"

    model = KMedoids(n_clusters=n_clusters, metric=metric)
    Xnaive = np.asarray([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    model.fit(Xnaive)

    assert_array_equal(
        model.cluster_centers_, [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    )
    assert_array_equal(model.labels_, [0, 1, 2])
    assert model.inertia_ == 0.0

    # diagonal must be zero, off-diagonals must be positive
    X_new = model.transform(Xnaive)
    for c in range(n_clusters):
        assert X_new[c, c] == 0
        for c2 in range(n_clusters):
            if c != c2:
                assert X_new[c, c2] > 0


def test_max_iter():
    """Test that warning message is thrown when max_iter is reached."""
    rng = np.random.RandomState(seed)
    X_iris = load_iris()["data"]

    model = KMedoids(
        n_clusters=10, init="random", random_state=rng, max_iter=1
    )
    msg = "Maximum number of iteration reached before"

    with pytest.warns(UserWarning, match=msg):
        model.fit(X_iris)


def test_kmedoids_iris():
    """Test kmedoids on the Iris dataset"""
    rng = np.random.RandomState(seed)
    X_iris = load_iris()["data"]

    ref_model = KMeans(n_clusters=3).fit(X_iris)

    avg_dist_to_closest_centroid = (
        ref_model.transform(X_iris).min(axis=1).mean()
    )

    for init in ["random", "heuristic", "k-medoids++"]:
        distance_metric = "euclidean"
        model = KMedoids(
            n_clusters=3, metric=distance_metric, init=init, random_state=rng
        )
        model.fit(X_iris)

        # test convergence in reasonable number of steps
        assert model.n_iter_ < (len(X_iris) // 10)

        distances = PAIRWISE_DISTANCE_FUNCTIONS[distance_metric](X_iris)
        avg_dist_to_random_medoid = np.mean(distances.ravel())
        avg_dist_to_closest_medoid = model.inertia_ / X_iris.shape[0]
        # We want distance-to-closest-medoid to be reduced from average
        # distance by more than 50%
        assert avg_dist_to_random_medoid > 2 * avg_dist_to_closest_medoid
        # When K-Medoids is using Euclidean distance,
        # we can compare its performance to
        # K-Means. We want the average distance to cluster centers
        # to be similar between K-Means and K-Medoids
        assert_allclose(
            avg_dist_to_closest_medoid, avg_dist_to_closest_centroid, rtol=0.1
        )


def test_kmedoids_fit_predict_transform():
    rng = np.random.RandomState(seed)
    model = KMedoids(random_state=rng)

    labels1 = model.fit_predict(X)
    assert len(labels1) == 100
    assert_array_equal(labels1, model.labels_)

    labels2 = model.predict(X)
    assert_array_equal(labels1, labels2)

    Xt1 = model.fit_transform(X)
    assert_array_equal(Xt1.shape, (100, model.n_clusters))

    Xt2 = model.transform(X)
    assert_array_equal(Xt1, Xt2)


def test_callable_distance_metric():
    rng = np.random.RandomState(seed)

    def my_metric(a, b):
        return np.sqrt(np.sum(np.power(a - b, 2)))

    model = KMedoids(random_state=rng, metric=my_metric)
    labels1 = model.fit_predict(X)
    assert len(labels1) == 100
    assert_array_equal(labels1, model.labels_)


def test_outlier_robustness():
    rng = np.random.RandomState(seed)
    kmeans = KMeans(n_clusters=2, random_state=rng)
    kmedoids = KMedoids(n_clusters=2, random_state=rng)

    X = [[-11, 0], [-10, 0], [-9, 0], [0, 0], [1, 0], [2, 0], [1000, 0]]

    kmeans.fit(X)
    kmedoids.fit(X)

    assert_array_equal(kmeans.labels_, [0, 0, 0, 0, 0, 0, 1])
    assert_array_equal(kmedoids.labels_, [0, 0, 0, 1, 1, 1, 1])


def test_kmedoids_on_sparse_input():
    rng = np.random.RandomState(seed)
    model = KMedoids(n_clusters=2, random_state=rng)
    row = np.array([1, 0])
    col = np.array([0, 4])
    data = np.array([1, 1])
    X = csc_matrix((data, (row, col)), shape=(2, 5))
    labels = model.fit_predict(X)
    assert len(labels) == 2
    assert_array_equal(labels, model.labels_)


# Test the build initialization.
def test_build():
    X, y = fetch_20newsgroups_vectorized(return_X_y=True)
    # Select only the first 500 samples
    X = X[:500]
    y = y[:500]
    # Precompute cosine distance matrix
    diss = cosine_distances(X)
    # run build
    ske = KMedoids(20, "precomputed", init="build", max_iter=0)
    ske.fit(diss)
    assert ske.inertia_ <= 230
    assert len(np.unique(ske.labels_)) == 20


def test_clara_consistency_iris():
    # test that CLARA is PAM when full sample is used

    rng = np.random.RandomState(seed)
    X_iris = load_iris()["data"]

    clara = CLARA(
        n_clusters=3,
        n_sampling_iter=1,
        n_sampling=len(X_iris),
        random_state=rng,
    )

    model = KMedoids(n_clusters=3, init="build", random_state=rng)

    model.fit(X_iris)
    clara.fit(X_iris)
    assert np.sum(model.labels_ == clara.labels_) == len(X_iris)


def test_seuclidean():
    with pytest.warns(None) as record:
        km = KMedoids(2, metric="seuclidean", method="pam")
        km.fit(np.array([0, 0, 0, 1]).reshape((4, 1)))
        km.predict(np.array([0, 0, 0, 1]).reshape((4, 1)))
        km.transform(np.array([0, 0, 0, 1]).reshape((4, 1)))
    assert len(record) == 0


def test_medoids_indices():
    rng = np.random.RandomState(seed)
    X_iris = load_iris()["data"]

    clara = CLARA(
        n_clusters=3,
        n_sampling_iter=1,
        n_sampling=len(X_iris),
        random_state=rng,
    )

    model = KMedoids(n_clusters=3, init="build", random_state=rng)

    centroids = np.array([X_iris[0], X_iris[50]])
    array_like_model = KMedoids(
        n_clusters=len(centroids), init=centroids, max_iter=0
    )

    model.fit(X_iris)
    clara.fit(X_iris)
    array_like_model.fit(X_iris)
    assert_array_equal(X_iris[model.medoid_indices_], model.cluster_centers_)
    assert_array_equal(X_iris[clara.medoid_indices_], clara.cluster_centers_)
    assert_array_equal(centroids, array_like_model.cluster_centers_)


def test_array_like_init():
    centroids = np.array([X_cc[0], X_cc[50]])

    expected = np.hstack([np.zeros(50), np.ones(50)])
    km = KMedoids(n_clusters=len(centroids), init=centroids)
    km.fit(X_cc)
    # # This test use data that are not perfectly separable so the
    # # accuracy is not 1. Accuracy around 0.85
    assert (np.mean(km.labels_ == expected) > 0.8) or (
        1 - np.mean(km.labels_ == expected) > 0.8
    )

    # Override n_clusters if array-like init method is used
    km = KMedoids(n_clusters=len(centroids) + 2, init=centroids)
    km.fit(X_cc)

    assert len(km.cluster_centers_) == len(centroids)



"""
==========================================
Common-nearest-neighbor clustering demo II
==========================================

Common-nearest neighbor clustering of data points following a density
criterion. Two points will be part of the same cluster if they share a
minimum number of common neighbors. Read more in the :ref:`User Guide
<commonnn>`. Compare this example to the results for
`other <https://scikit-learn.org/stable/auto_examples/cluster/plot_cluster_comparison.html>`_
cluster algorithms.

"""

import matplotlib.pyplot as plt
import numpy as np

from sklearn_extra.cluster import CommonNNClustering
from sklearn import datasets
from sklearn.preprocessing import StandardScaler


print(__doc__)


np.random.seed(42)
n = 2000

# circles
circles, _ = datasets.make_circles(
    n_samples=n, factor=0.5, noise=0.05, random_state=10
)

circles = StandardScaler().fit_transform(circles)

# blobs
blobs, _ = datasets.make_blobs(
    centers=[[-9, -8], [11, -10], [12, 12]], n_samples=n, random_state=10
)

blobs = StandardScaler().fit_transform(blobs)

# moons
moons, _ = datasets.make_moons(n_samples=n, noise=0.05, random_state=10)

moons = StandardScaler().fit_transform(moons)

# no_structure
no_structure = np.random.rand(n, 2)
no_structure = StandardScaler().fit_transform(no_structure)

# aniso
X, y = datasets.make_blobs(n_samples=n, random_state=170)

transformation = [[0.6, -0.6], [-0.4, 0.8]]
aniso = np.dot(X, transformation)
aniso = StandardScaler().fit_transform(aniso)

# varied
varied, _ = datasets.make_blobs(
    n_samples=n, cluster_std=[1.0, 2, 0.5], random_state=170
)

varied = StandardScaler().fit_transform(varied)

fits = [
    ("circles", circles, {"eps": 0.2, "min_samples": 5}),
    ("moons", moons, {"eps": 0.2, "min_samples": 5}),
    ("varied", varied, {"eps": 0.2, "min_samples": 15}),
    ("aniso", aniso, {"eps": 0.18, "min_samples": 12}),
    ("blobs", blobs, {"eps": 0.2, "min_samples": 5}),
    ("none", no_structure, {"eps": 0.2, "min_samples": 5}),
]

fig, ax = plt.subplots(2, 3)
ax = ax.flatten()
for index, (name, data, params) in enumerate(fits):
    cobj = CommonNNClustering(**params).fit(data)
    labels = cobj.labels_
    ax[index].plot(
        *data[np.where(labels == -1)[0]].T,
        linestyle="",
        color="None",
        marker="o",
        markersize=4,
        markerfacecolor="gray",
        markeredgecolor="k",
    )

    for cluster_number in range(0, int(np.max(labels)) + 1):
        ax[index].plot(
            *data[np.where(labels == cluster_number)[0]].T,
            linestyle="",
            marker="o",
            markersize=4,
            markeredgecolor="k",
        )

    ax[index].set(
        **{
            "xlabel": None,
            "ylabel": None,
            "xlim": (-2.5, 2.5),
            "ylim": (-2.5, 2.5),
            "xticks": (),
            "yticks": (),
            "aspect": "equal",
            "title": name,
        }
    )



# -*- coding: utf-8 -*-
"""
================================================================
A demo of Robust Regression on real dataset "california housing"
================================================================
In this example we compare the RobustWeightedRegressor to other scikit-learn
regressors on the real dataset california housing.

One of the main point of this example is the importance of taking into account
outliers in the test dataset when dealing with real datasets.

For this example, we took a parameter so that RobustWeightedRegressor is better
than RANSAC and TheilSen when talking about the mean squared error and it
is better than the SGDRegressor when talking about the median squared error.
Depending on what criterion one want to optimize, the parameter measuring
robustness in RobustWeightedRegressor can change and this is not so
straightforward when using RANSAC and TheilSenRegressor.
"""
import matplotlib.pyplot as plt
import numpy as np
from sklearn_extra.robust import RobustWeightedRegressor
from sklearn.linear_model import (
    SGDRegressor,
    TheilSenRegressor,
    RANSACRegressor,
)
from sklearn.datasets import fetch_california_housing
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import RobustScaler


def quadratic_loss(est, X, y, X_test, y_test):
    est.fit(X, y)
    return (est.predict(X_test) - y_test) ** 2


X, y = fetch_california_housing(as_frame=False, return_X_y=True)
# Sub-sample for faster computation.
X = X[:1000]
y = y[:1000]

# Scale the dataset with sklearn RobustScaler (important for this algorithm)
X = RobustScaler().fit_transform(X)

# Using GridSearchCV, we do a light tuning of the parameters for SGDRegressor
# and RobustWeightedEstimator. A fine tune is possible but not necessary to
# illustrate the problem of outliers in the output.
estimators = [
    (
        "SGD",
        SGDRegressor(learning_rate="adaptive", eta0=1e-2),
    ),
    (
        "RobustWeightedRegressor",
        RobustWeightedRegressor(
            weighting="huber",
            c=0.1,
            eta0=1e-2,
            sgd_args={
                "learning_rate": "invscaling",
            },
        ),
    ),
    ("RANSAC", RANSACRegressor()),
    ("TheilSen", TheilSenRegressor()),
]

M = 10
res = np.zeros(shape=[len(estimators), M, 2])

for f in range(M):
    print("\r Progress: %s / %s" % (f + 1, M), end="")

    rng = np.random.RandomState(f)

    # Split in a training set and a test set
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.2, random_state=rng
    )

    for i, (name, est) in enumerate(estimators):
        cv = quadratic_loss(est, X_train, y_train, X_test, y_test)

        # It is preferable to use the median of the validation losses
        # because it is possible that some outliers are present in the
        # test set.
        # We compute both for comparison.
        res[i, f, 0] = np.mean(cv)
        res[i, f, 1] = np.median(cv)

fig, (axe1, axe2) = plt.subplots(2, 1)
names = [name for name, est in estimators]

axe1.boxplot(res[:, :, 0].T, labels=names)

axe2.boxplot(res[:, :, 1].T, labels=names)


axe1.set_title("mean of errors")
axe2.set_title("median of errors")

fig.suptitle("Boxplots of the test squared error")



from sklearn_extra.robust.robust_weighted_estimator import (
    RobustWeightedClassifier,
    RobustWeightedKMeans,
    RobustWeightedRegressor,
)

__all__ = [
    "RobustWeightedClassifier",
    "RobustWeightedKMeans",
    "RobustWeightedRegressor",
]



import pytest
import numpy as np

from numpy.testing import assert_array_almost_equal
from sklearn.metrics.pairwise import rbf_kernel

from sklearn_extra.kernel_approximation import Fastfood


# generate data
rng = np.random.RandomState(0)
X = rng.random_sample(size=(300, 50))
Y = rng.random_sample(size=(300, 50))
X /= X.sum(axis=1)[:, np.newaxis]
Y /= Y.sum(axis=1)[:, np.newaxis]


@pytest.mark.parametrize(
    "message, input_, expected",
    [
        ("test n is scaled to be a multiple of d", (16, 20), (16, 32, 2)),
        ("test n equals d", (16, 16), (16, 16, 1)),
        ("test n becomes power of two", (3, 16), (4, 16, 4)),
        ("test all", (7, 12), (8, 16, 2)),
    ],
)
def test_fastfood_enforce_dimensionality_constraint(message, input_, expected):
    d, n = input_
    output = Fastfood._enforce_dimensionality_constraints(d, n)
    assert expected == output, message


def test_fastfood():
    """test that Fastfood fast approximates kernel on random data"""
    # compute exact kernel
    gamma = 10.0
    kernel = rbf_kernel(X, Y, gamma=gamma)

    sigma = np.sqrt(1 / (2 * gamma))

    # approximate kernel mapping
    ff_transform = Fastfood(sigma, n_components=1000, random_state=42)

    pars = ff_transform.fit(X)
    X_trans = pars.transform(X)
    Y_trans = ff_transform.transform(Y)

    kernel_approx = np.dot(X_trans, Y_trans.T)

    print("approximation:", kernel_approx[:5, :5])
    print("true kernel:", kernel[:5, :5])
    assert_array_almost_equal(kernel, kernel_approx, decimal=1)



import pytest
from sklearn.utils import estimator_checks

from sklearn_extra.kernel_approximation import Fastfood
from sklearn_extra.kernel_methods import EigenProClassifier, EigenProRegressor
from sklearn_extra.cluster import KMedoids, CommonNNClustering, CLARA
from sklearn_extra.robust import (
    RobustWeightedClassifier,
    RobustWeightedRegressor,
    RobustWeightedKMeans,
)


ALL_ESTIMATORS = [
    Fastfood,
    KMedoids,
    CLARA,
    EigenProClassifier,
    EigenProRegressor,
    CommonNNClustering,
    RobustWeightedKMeans,
    RobustWeightedRegressor,
    RobustWeightedClassifier,
]


@estimator_checks.parametrize_with_checks([cls() for cls in ALL_ESTIMATORS])
def test_all_estimators(estimator, check, request):
    # TODO: fix this common test failure cf #41
    if isinstance(
        estimator, EigenProClassifier
    ) and "function check_classifier_multioutput" in str(check):
        request.applymarker(
            pytest.mark.xfail(run=False, reason="See issue #41")
        )

    return check(estimator)



import numpy as np
import numpy.testing as npt
from scipy.linalg import hadamard
import pytest

from sklearn_extra.utils._cyfht import fht as cyfht
from sklearn_extra.utils._cyfht import fht2 as cyfht2


def test_wikipedia_example():
    input_ = np.array([1, 0, 1, 0, 0, 1, 1, 0], dtype=np.float64)
    copy = input_.copy()
    H = hadamard(8)
    cyfht(input_)
    npt.assert_array_equal(np.dot(copy, H), input_)


def test_numerical_fuzzing_fht():
    for length in [2, 4, 8, 16, 32, 64]:
        input_ = np.random.normal(size=length)
        copy = input_.copy()
        H = hadamard(length)
        cyfht(input_)
        npt.assert_array_almost_equal(np.dot(copy, H), input_)


def test_numerical_fuzzing_fht2():
    for length in [2, 4, 8, 16, 32, 64]:
        for rows in [1, 2, 3, 4, 5]:
            input_ = np.random.normal(size=(rows, length))
            copy = input_.copy()
            H = hadamard(length)
            cyfht2(input_)
            npt.assert_array_almost_equal(np.dot(copy, H), input_)


def test_exception_when_input_not_power_two():
    msg = "Length of input for fht must be a power of two"
    with pytest.raises(ValueError, match=msg):
        cyfht(np.zeros(9, dtype=np.float64))
    msg = "Length of rows for fht2 must be a power of two"
    with pytest.raises(ValueError, match=msg):
        cyfht2(np.zeros((2, 9), dtype=np.float64))



# -*- coding: utf-8 -*-
"""
=============================================================
Robust regression on simulated corrupted dataset
=============================================================
In this example we compare the RobustWeightedRegressor
with various robust regression algorithms from scikit-learn.
"""
import matplotlib.pyplot as plt
import numpy as np

from sklearn_extra.robust import RobustWeightedRegressor
from sklearn.utils import shuffle
from sklearn.linear_model import (
    SGDRegressor,
    LinearRegression,
    TheilSenRegressor,
    RANSACRegressor,
    HuberRegressor,
)

# Sample along a line with a Gaussian noise.
rng = np.random.RandomState(42)
X = rng.uniform(-1, 1, size=[100])
y = X + 0.1 * rng.normal(size=100)

# Change the 5 last entries to an outlier.
X[-5:] = 10
X = X.reshape(-1, 1)
y[-5:] = -1

# Shuffle the data so that we don't know where the outlier is.
X, y = shuffle(X, y, random_state=rng)
estimators = [
    ("OLS", LinearRegression()),
    ("Theil-Sen", TheilSenRegressor(random_state=rng)),
    ("RANSAC", RANSACRegressor(random_state=rng)),
    ("HuberRegressor", HuberRegressor()),
    (
        "SGD epsilon loss",
        SGDRegressor(loss="epsilon_insensitive", random_state=rng),
    ),
    (
        "RobustWeightedRegressor",
        RobustWeightedRegressor(weighting="mom", k=7, random_state=rng),
        # The parameter k is set larger to the number of outliers
        # because here we know it.
    ),
]


colors = {
    "OLS": "turquoise",
    "Theil-Sen": "gold",
    "RANSAC": "lightgreen",
    "HuberRegressor": "black",
    "RobustWeightedRegressor": "magenta",
    "SGD epsilon loss": "purple",
}
linestyle = {
    "OLS": "-",
    "SGD epsilon loss": "-",
    "Theil-Sen": "-.",
    "RANSAC": "--",
    "HuberRegressor": "--",
    "RobustWeightedRegressor": "--",
}
lw = 3

x_plot = np.linspace(X.min(), X.max())

plt.plot(X, y, "b+")

for name, estimator in estimators:
    estimator.fit(X, y)
    y_plot = estimator.predict(x_plot[:, np.newaxis])
    plt.plot(
        x_plot,
        y_plot,
        color=colors[name],
        linestyle=linestyle[name],
        linewidth=lw,
        label="%s" % (name),
    )

legend = plt.legend(loc="upper right", prop=dict(size="x-small"))

plt.title(
    "Scatter plot of training set and representation of"
    " estimation functions"
)
plt.show()



# -*- coding: utf-8 -*-
"""
==============================================================
A demo of Robust Classification on Simulated corrupted dataset
==============================================================
In this example we compare the RobustWeightedClassifier using SGDClassifier
for classification with the vanilla SGDClassifier with various losses.
"""
import matplotlib.pyplot as plt
import numpy as np
from sklearn_extra.robust import RobustWeightedClassifier
from sklearn.linear_model import SGDClassifier
from sklearn.datasets import make_blobs
from sklearn.utils import shuffle


rng = np.random.RandomState(42)

# Sample two Gaussian blobs
X, y = make_blobs(
    n_samples=100, centers=np.array([[-1, -1], [1, 1]]), random_state=rng
)

# Change the first 3 entries to outliers
for f in range(3):
    X[f] = [20, 3] + rng.normal(size=2) * 0.1
    y[f] = 0

# Shuffle the data so that we don't know where the outlier is.
X, y = shuffle(X, y, random_state=rng)

estimators = [
    (
        "SGDClassifier, Hinge loss",
        SGDClassifier(loss="hinge", random_state=rng),
    ),
    ("SGDClassifier, log loss", SGDClassifier(loss="log", random_state=rng)),
    (
        "SGDClassifier, modified_huber loss",
        SGDClassifier(loss="modified_huber", random_state=rng),
    ),
    (
        "RobustWeightedClassifier",
        RobustWeightedClassifier(
            max_iter=100,
            weighting="mom",
            k=8,
            random_state=rng,
        ),
        # The parameter k is set larger the number of outliers
        # because here we know it. max_iter is set to 100. One may want
        # to play with the number of iteration or the optimization scheme of
        # the base_estimator to get good results.
    ),
]


# Helping function to represent estimators
def plot_classif(clf, X, y, ax):
    x_min, x_max = X[:, 0].min() - 0.5, X[:, 0].max() + 0.5
    y_min, y_max = X[:, 1].min() - 0.5, X[:, 1].max() + 0.5
    h = 0.02  # step size in the mesh
    xx, yy = np.meshgrid(
        np.arange(x_min, x_max, h), np.arange(y_min, y_max, h)
    )
    Z = clf.predict(np.c_[xx.ravel(), yy.ravel()])

    # Put the result into a color plot
    Z = Z.reshape(xx.shape)
    ax.pcolormesh(xx, yy, Z, cmap=plt.cm.Paired)
    ax.scatter(X[:, 0], X[:, 1], c=y)


fig, axes = plt.subplots(2, 2)

for i, (name, estimator) in enumerate(estimators):
    ax = axes.flat[i]
    estimator.fit(X, y)
    plot_classif(estimator, X, y, ax)
    ax.set_title(name)
    ax.axis("off")

fig.suptitle(
    "Scatter plot of training set and representation of"
    " estimation functions"
)
plt.show()



#! /usr/bin/env python
import os

from setuptools import find_packages, setup, Extension

import numpy as np

from Cython.Build import cythonize
from Cython.Distutils import build_ext

# get __version__ from _version.py
ver_file = os.path.join("sklearn_extra", "_version.py")
with open(ver_file) as f:
    exec(f.read())

DISTNAME = "sklearn-kmedoids-on-string"
DESCRIPTION = "A set of tools for scikit-learn."
with open("README.rst", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()
URL = "https://github.com/mln-wave/scikit-learn-extra"
LICENSE = "new BSD"
DOWNLOAD_URL = "https://github.com/mln-wave/scikit-learn-extra"
VERSION = __version__  # noqa
INSTALL_REQUIRES = ["numpy>=1.13.3", "scipy>=0.19.1", "scikit-learn>=0.23.0"]
CLASSIFIERS = [
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "License :: OSI Approved",
    "Programming Language :: Python",
    "Topic :: Software Development",
    "Topic :: Scientific/Engineering",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: MacOS",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: Implementation :: CPython",
]
EXTRAS_REQUIRE = {
    "tests": ["pytest", "pytest-cov"],
    "docs": [
        "pillow",
        "sphinx",
        "sphinx-gallery",
        "sphinx_rtd_theme",
        "numpydoc",
        "matplotlib",
    ],
}
libraries = []
if os.name == "posix":
    libraries.append("m")

args = {
    "ext_modules": cythonize(
        [
            Extension(
                "sklearn_extra.utils._cyfht",
                ["sklearn_extra/utils/_cyfht.pyx"],
                include_dirs=[np.get_include()],
            ),
            Extension(
                "sklearn_extra.cluster._k_medoids_helper",
                ["sklearn_extra/cluster/_k_medoids_helper.pyx"],
                include_dirs=[np.get_include()],
            ),
            Extension(
                "sklearn_extra.robust._robust_weighted_estimator_helper",
                ["sklearn_extra/robust/_robust_weighted_estimator_helper.pyx"],
                include_dirs=[np.get_include()],
                libraries=libraries,
            ),
            Extension(
                "sklearn_extra.cluster._commonnn_inner",
                ["sklearn_extra/cluster/_commonnn_inner.pyx"],
                include_dirs=[np.get_include()],
                language="c++",
            ),
        ]
    ),
    "cmdclass": dict(build_ext=build_ext),
}


setup(
    name=DISTNAME,
    description=DESCRIPTION,
    long_description_content_type="text/x-rst",
    license=LICENSE,
    url=URL,
    version=VERSION,
    download_url=DOWNLOAD_URL,
    long_description=LONG_DESCRIPTION,
    zip_safe=False,  # the package can run out of an .egg file
    classifiers=CLASSIFIERS,
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    python_requires=">=3.6",
    **args
)



# cython: infer_types=True
# Fast swap step in PAM algorithm for k_medoid.
# Author: Timothée Mathieu
# License: 3-clause BSD

cimport cython
import numpy as np
cimport numpy as np

from sklearn.utils.extmath import row_norms
from cython cimport floating

import sys
from time import time

from libc.math cimport exp, log, sqrt, pow, fabs
cimport numpy as np
from numpy.math cimport INFINITY


# Modified from sklearn.cluster._k_means_fast.pyx
np.import_array()

cdef floating _euclidean_dense_dense(
        floating* a,  # IN
        floating* b,  # IN
        int n_features) nogil:
    """Euclidean distance between a dense and b dense"""
    cdef:
        int i
        int n = n_features // 4
        int rem = n_features % 4
        floating result = 0

    # We manually unroll the loop for better cache optimization.
    for i in range(n):
        result += ((a[0] - b[0]) * (a[0] - b[0])
                  +(a[1] - b[1]) * (a[1] - b[1])
                  +(a[2] - b[2]) * (a[2] - b[2])
                  +(a[3] - b[3]) * (a[3] - b[3]))
        a += 4; b += 4

    for i in range(rem):
        result += (a[i] - b[i]) * (a[i] - b[i])

    return result



cpdef np.ndarray[floating] _kmeans_loss(np.ndarray[floating, ndim=2, mode='c'] X,
                                        int[:] labels):
    """Compute inertia

    squared distancez between each sample and its assigned center.
    """
    if floating is float:
        dtype = np.float32
    elif floating is double:
        dtype = np.double

    cdef:
        int n_samples = X.shape[0]
        int n_features = X.shape[1]
        int i, j
        int n_classes = len(np.unique(labels))
        np.ndarray[floating, ndim=2] centers = np.zeros([n_classes,
                                                         n_features],
                                                         dtype = dtype)
        np.ndarray[long] num_in_cluster = np.zeros(n_classes, dtype = int)
        np.ndarray[floating] inertias = np.zeros(n_samples, dtype = dtype)
    for i in range(n_samples):
        for j in range(n_features):
            centers[labels[i], j] += X[i, j]
        num_in_cluster[labels[i]] += 1

    for i in range(n_classes):
        for j in range(n_features):
            centers[i, j] /= num_in_cluster[i]

    for i in range(n_samples):
        j = labels[i]
        inertias[i] = _euclidean_dense_dense(&X[i, 0], &centers[j, 0], n_features)
    return inertias





# Regression and Classification losses, from scikit-learn.




# ----------------------------------------
# Extension Types for Loss Functions
# ----------------------------------------

cdef class LossFunction:
    """Base class for convex loss functions"""

    cdef double loss(self, double p, double y) nogil:
        """Evaluate the loss function.

        Parameters
        ----------
        p : double
            The prediction, p = w^T x
        y : double
            The true value (aka target)

        Returns
        -------
        double
            The loss evaluated at `p` and `y`.
        """
        return 0.

    def py_dloss(self, double p, double y):
        """Python version of `dloss` for testing.

        Pytest needs a python function and can't use cdef functions.
        """
        return self.dloss(p, y)

    def py_loss(self, double p, double y):
        """Python version of `dloss` for testing.

        Pytest needs a python function and can't use cdef functions.
        """
        return self.loss(p, y)


    cdef double dloss(self, double p, double y) nogil:
        """Evaluate the derivative of the loss function with respect to
        the prediction `p`.

        Parameters
        ----------
        p : double
            The prediction, p = w^T x
        y : double
            The true value (aka target)
        Returns
        -------
        double
            The derivative of the loss function with regards to `p`.
        """
        return 0.


cdef class Regression(LossFunction):
    """Base class for loss functions for regression"""

    cdef double loss(self, double p, double y) nogil:
        return 0.

    cdef double dloss(self, double p, double y) nogil:
        return 0.


cdef class Classification(LossFunction):
    """Base class for loss functions for classification"""

    cdef double loss(self, double p, double y) nogil:
        return 0.

    cdef double dloss(self, double p, double y) nogil:
        return 0.


cdef class ModifiedHuber(Classification):
    """Modified Huber loss for binary classification with y in {-1, 1}

    This is equivalent to quadratically smoothed SVM with gamma = 2.

    See T. Zhang 'Solving Large Scale Linear Prediction Problems Using
    Stochastic Gradient Descent', ICML'04.
    """
    cdef double loss(self, double p, double y) nogil:
        cdef double z = p * y
        if z >= 1.0:
            return 0.0
        elif z >= -1.0:
            return (1.0 - z) * (1.0 - z)
        else:
            return -4.0 * z

    cdef double dloss(self, double p, double y) nogil:
        cdef double z = p * y
        if z >= 1.0:
            return 0.0
        elif z >= -1.0:
            return 2.0 * (1.0 - z) * -y
        else:
            return -4.0 * y

    def __reduce__(self):
        return ModifiedHuber, ()


cdef class Hinge(Classification):
    """Hinge loss for binary classification tasks with y in {-1,1}

    Parameters
    ----------

    threshold : float > 0.0
        Margin threshold. When threshold=1.0, one gets the loss used by SVM.
        When threshold=0.0, one gets the loss used by the Perceptron.
    """

    cdef double threshold

    def __init__(self, double threshold=1.0):
        self.threshold = threshold

    cdef double loss(self, double p, double y) nogil:
        cdef double z = p * y
        if z <= self.threshold:
            return self.threshold - z
        return 0.0

    cdef double dloss(self, double p, double y) nogil:
        cdef double z = p * y
        if z <= self.threshold:
            return -y
        return 0.0

    def __reduce__(self):
        return Hinge, (self.threshold,)


cdef class SquaredHinge(Classification):
    """Squared Hinge loss for binary classification tasks with y in {-1,1}

    Parameters
    ----------

    threshold : float > 0.0
        Margin threshold. When threshold=1.0, one gets the loss used by
        (quadratically penalized) SVM.
    """

    cdef double threshold

    def __init__(self, double threshold=1.0):
        self.threshold = threshold

    cdef double loss(self, double p, double y) nogil:
        cdef double z = self.threshold - p * y
        if z > 0:
            return z * z
        return 0.0

    cdef double dloss(self, double p, double y) nogil:
        cdef double z = self.threshold - p * y
        if z > 0:
            return -2 * y * z
        return 0.0

    def __reduce__(self):
        return SquaredHinge, (self.threshold,)


cdef class Log(Classification):
    """Logistic regression loss for binary classification with y in {-1, 1}"""

    cdef double loss(self, double p, double y) nogil:
        cdef double z = p * y
        # approximately equal and saves the computation of the log
        if z > 18:
            return exp(-z)
        if z < -18:
            return -z
        return log(1.0 + exp(-z))

    cdef double dloss(self, double p, double y) nogil:
        cdef double z = p * y
        # approximately equal and saves the computation of the log
        if z > 18.0:
            return exp(-z) * -y
        if z < -18.0:
            return -y
        return -y / (exp(z) + 1.0)

    def __reduce__(self):
        return Log, ()


cdef class SquaredLoss(Regression):
    """Squared loss traditional used in linear regression."""
    cdef double loss(self, double p, double y) nogil:
        return 0.5 * (p - y) * (p - y)

    cdef double dloss(self, double p, double y) nogil:
        return p - y

    def __reduce__(self):
        return SquaredLoss, ()


cdef class Huber(Regression):
    """Huber regression loss

    Variant of the SquaredLoss that is robust to outliers (quadratic near zero,
    linear in for large errors).

    https://en.wikipedia.org/wiki/Huber_Loss_Function
    """

    cdef double c

    def __init__(self, double c):
        self.c = c

    cdef double loss(self, double p, double y) nogil:
        cdef double r = p - y
        cdef double abs_r = fabs(r)
        if abs_r <= self.c:
            return 0.5 * r * r
        else:
            return self.c * abs_r - (0.5 * self.c * self.c)

    cdef double dloss(self, double p, double y) nogil:
        cdef double r = p - y
        cdef double abs_r = fabs(r)
        if abs_r <= self.c:
            return r
        elif r > 0.0:
            return self.c
        else:
            return -self.c

    def __reduce__(self):
        return Huber, (self.c,)



# License: BSD 3 clause

from math import sqrt
import numpy as np
from scipy.stats import chi

from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.utils import check_array, check_random_state

from ..utils._cyfht import fht2 as cyfht


class Fastfood(BaseEstimator, TransformerMixin):
    """Approximates feature map of an RBF kernel by Monte Carlo approximation
    of its Fourier transform.

    Fastfood replaces the random matrix of Random Kitchen Sinks (RBFSampler)
    with an approximation that uses the Walsh-Hadamard transformation to gain
    significant speed and storage advantages.  The computational complexity for
    mapping a single example is O(n_components log d).  The space complexity is
    O(n_components).  Hint: n_components should be a power of two. If this is
    not the case, the next higher number that fulfills this constraint is
    chosen automatically.

    Parameters
    ----------
    sigma : float
        Parameter of RBF kernel: exp(-(1/(2*sigma^2)) * x^2)

    n_components : int
        Number of Monte Carlo samples per original feature.
        Equals the dimensionality of the computed feature space.

    tradeoff_mem_accuracy : "accuracy" or "mem", default: 'accuracy'
        mem:        This version is not as accurate as the option "accuracy",
                    but is consuming less memory.
        accuracy:   The final feature space is of dimension 2*n_components,
                    while being more accurate and consuming more memory.

    random_state : {int, RandomState}, optional
        If int, random_state is the seed used by the random number generator;
        if RandomState instance, random_state is the random number generator.

    Notes
    -----
    See "Fastfood | Approximating Kernel Expansions in Loglinear Time" by
    Quoc Le, Tamas Sarl and Alex Smola.

    Examples
    --------
    See scikit-learn-fastfood/examples/plot_digits_classification_fastfood.py
    for an example how to use fastfood with a primal classifier in comparison
    to an usual rbf-kernel with a dual classifier.

    """

    def __init__(
        self,
        sigma=sqrt(1 / 2),
        n_components=100,
        tradeoff_mem_accuracy="accuracy",
        random_state=None,
    ):
        self.sigma = sigma
        self.n_components = n_components
        self.random_state = random_state
        # map to 2*n_components features or to n_components features with less
        # accuracy
        self.tradeoff_mem_accuracy = tradeoff_mem_accuracy

    @staticmethod
    def _is_number_power_of_two(n):
        return n != 0 and ((n & (n - 1)) == 0)

    @staticmethod
    def _enforce_dimensionality_constraints(d, n):
        if not (Fastfood._is_number_power_of_two(d)):
            # find d that fulfills 2^l
            d = np.power(2, np.floor(np.log2(d)) + 1)
        divisor, remainder = divmod(n, d)
        times_to_stack_v = int(divisor)
        if remainder != 0:
            # output info, that we increase n so that d is a divider of n
            n = (divisor + 1) * d
            times_to_stack_v = int(divisor + 1)
        return int(d), int(n), times_to_stack_v

    def _pad_with_zeros(self, X):
        try:
            X_padded = np.pad(
                X,
                ((0, 0), (0, self._number_of_features_to_pad_with_zeros)),
                "constant",
            )
        except AttributeError:
            zeros = np.zeros(
                (X.shape[0], self._number_of_features_to_pad_with_zeros)
            )
            X_padded = np.concatenate((X, zeros), axis=1)

        return X_padded

    @staticmethod
    def _approx_fourier_transformation_multi_dim(result):
        cyfht(result)

    @staticmethod
    def _l2norm_along_axis1(X):
        return np.sqrt(np.einsum("ij,ij->i", X, X))

    def _uniform_vector(self, rng):
        if self.tradeoff_mem_accuracy != "accuracy":
            return rng.uniform(0, 2 * np.pi, size=self._n)
        else:
            return None

    def _apply_approximate_gaussian_matrix(self, B, G, P, X):
        """Create mapping of all x_i by applying B, G and P step-wise"""
        num_examples = X.shape[0]

        result = np.multiply(B, X.reshape((1, num_examples, 1, self._d)))
        result = result.reshape(
            (num_examples * self._times_to_stack_v, self._d)
        )
        Fastfood._approx_fourier_transformation_multi_dim(result)
        result = result.reshape((num_examples, -1))
        np.take(result, P, axis=1, mode="wrap", out=result)
        np.multiply(
            np.ravel(G), result.reshape(num_examples, self._n), out=result
        )
        result = result.reshape(num_examples * self._times_to_stack_v, self._d)
        Fastfood._approx_fourier_transformation_multi_dim(result)
        return result

    def _scale_transformed_data(self, S, VX):
        """Scale mapped data VX to match kernel(e.g. RBF-Kernel)"""
        VX = VX.reshape(-1, self._times_to_stack_v * self._d)

        return (
            1 / (self.sigma * np.sqrt(self._d)) * np.multiply(np.ravel(S), VX)
        )

    def _phi(self, X):
        if self.tradeoff_mem_accuracy == "accuracy":
            return (1 / np.sqrt(X.shape[1])) * np.hstack(
                [np.cos(X), np.sin(X)]
            )
        else:
            np.cos(X + self._U, X)
            return X * np.sqrt(2.0 / X.shape[1])

    def fit(self, X, y=None):
        """Fit the model with X.

        Samples a couple of random based vectors to approximate a Gaussian
        random projection matrix to generate n_components features.

        Parameters
        ----------
        X : {array-like}, shape (n_samples, n_features)
            Training data, where n_samples in the number of samples
            and n_features is the number of features.

        Returns
        -------
        self : object
            Returns the transformer.
        """
        X = check_array(X, order="C", dtype=np.float64)

        d_orig = X.shape[1]
        rng = check_random_state(self.random_state)

        (
            self._d,
            self._n,
            self._times_to_stack_v,
        ) = Fastfood._enforce_dimensionality_constraints(
            d_orig, self.n_components
        )
        self._number_of_features_to_pad_with_zeros = self._d - d_orig

        self._G = rng.normal(size=(self._times_to_stack_v, self._d))
        self._B = rng.choice(
            [-1, 1], size=(self._times_to_stack_v, self._d), replace=True
        )
        self._P = np.hstack(
            [
                (i * self._d) + rng.permutation(self._d)
                for i in range(self._times_to_stack_v)
            ]
        )
        self._S = np.multiply(
            1 / self._l2norm_along_axis1(self._G).reshape((-1, 1)),
            chi.rvs(
                self._d,
                size=(self._times_to_stack_v, self._d),
                random_state=rng,
            ),
        )

        self._U = self._uniform_vector(rng)
        self.is_fitted_ = True

        return self

    def transform(self, X):
        """Apply the approximate feature map to X.

        Parameters
        ----------
        X : {array-like}, shape (n_samples, n_features)
            New data, where n_samples in the number of samples
            and n_features is the number of features.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
        """
        X = check_array(X, order="C", dtype=np.float64)
        X_padded = self._pad_with_zeros(X)
        HGPHBX = self._apply_approximate_gaussian_matrix(
            self._B, self._G, self._P, X_padded
        )
        VX = self._scale_transformed_data(self._S, HGPHBX)
        return self._phi(VX)



# Authors: Alex Li <7Alex7Li@gmail.com>
#          Siyuan Ma <Siyuan.ma9@gmail.com>

import numpy as np
from scipy.linalg import eigh, LinAlgError
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.metrics.pairwise import pairwise_kernels, euclidean_distances
from sklearn.utils import check_random_state
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import check_is_fitted, check_X_y


class BaseEigenPro(BaseEstimator):
    """
    Base class for EigenPro iteration.
    """

    def __init__(
        self,
        batch_size="auto",
        n_epoch=2,
        n_components=1000,
        subsample_size="auto",
        kernel="rbf",
        gamma="scale",
        degree=3,
        coef0=1,
        kernel_params=None,
        random_state=None,
    ):
        self.batch_size = batch_size
        self.n_epoch = n_epoch
        self.n_components = n_components
        self.subsample_size = subsample_size
        self.kernel = kernel
        self.gamma = gamma
        self.degree = degree
        self.coef0 = coef0
        self.kernel_params = kernel_params
        self.random_state = random_state

    def _kernel(self, X, Y):
        """Calculate the kernel matrix

        Parameters
        ---------
        X : {float, array}, shape = [n_samples, n_features]
            Input data.

        Y : {float, array}, shape = [n_centers, n_targets]
            Kernel centers.

        Returns
        -------
        K : {float, array}, shape = [n_samples, n_centers]
            Kernel matrix.
        """
        if (
            self.kernel != "rbf"
            and self.kernel != "laplace"
            and self.kernel != "cauchy"
        ):
            if callable(self.kernel):
                params = self.kernel_params or {}
            else:
                params = {
                    "gamma": self.gamma_,
                    "degree": self.degree,
                    "coef0": self.coef0,
                }
            return pairwise_kernels(
                X, Y, metric=self.kernel, filter_params=True, **params
            )
        distance = euclidean_distances(X, Y, squared=True)
        bandwidth = np.float32(1.0 / np.sqrt(2.0 * self.gamma_))
        if self.kernel == "rbf":
            distance = -self.gamma_ * distance
            K = np.exp(distance)
        elif self.kernel == "laplace":
            d = np.maximum(distance, 0)
            K = np.exp(-np.sqrt(d) / bandwidth)
        else:  # self.kernel == "cauchy":
            K = 1 / (1 + 2.0 * self.gamma_ * distance)
        return K

    def _nystrom_svd(self, X, n_components):
        """Compute the top eigensystem of a kernel
        operator using Nystrom method

        Parameters
        ----------
        X : {float, array}, shape = [n_subsamples, n_features]
            Subsample feature matrix.

        n_components : int
            Number of top eigencomponents to be restored.

        Returns
        -------
        E : {float, array}, shape = [k]
            Top eigenvalues.

        Lambda : {float, array}, shape = [n_subsamples, k]
            Top eigenvectors of a subsample kernel matrix (which can be
            directly used to approximate the eigenfunctions of the kernel
            operator).
        """
        m, _ = X.shape
        K = self._kernel(X, X)

        W = K / m
        try:
            E, Lambda = eigh(W, eigvals=(m - n_components, m - 1))
        except LinAlgError:
            # Use float64 when eigh fails due to precision
            W = np.float64(W)
            E, Lambda = eigh(W, eigvals=(m - n_components, m - 1))
            E, Lambda = np.float32(E), np.float32(Lambda)
        # Flip so eigenvalues are in descending order.
        E = np.maximum(np.float32(1e-7), np.flipud(E))
        Lambda = np.fliplr(Lambda)[:, :n_components] / np.sqrt(
            m, dtype="float32"
        )

        return E, Lambda

    def _setup(self, feat, max_components, mG, alpha):
        """Compute preconditioner and scale factors for EigenPro iteration

        Parameters
        ----------
        feat : {float, array}, shape = [n_samples, n_features]
            Feature matrix (normally from training data).

        max_components : int
            Maximum number of components to be used in EigenPro iteration.

        mG : int
            Maximum batch size to fit in memory.

        alpha : float
            Exponential factor (< 1) for eigenvalue ratio.

        Returns
        -------
        max_S : float
            Normalized largest eigenvalue.

        max_kxx : float
            Maximum of k(x,x) where k is the EigenPro kernel.

        E : {float, array}, shape = [k]
            Preconditioner for EigenPro

        Lambda : {float, array}, shape = [n_subsamples, k]
            Top eigenvectors of a subsample kernel matrix
        """
        alpha = np.float32(alpha)

        # Estimate eigenvalues (S) and eigenvectors (V) of the kernel matrix
        # corresponding to the feature matrix.
        E, Lambda = self._nystrom_svd(feat, max_components)
        n_subsamples = feat.shape[0]

        # Calculate the number of components to be used such that the
        # corresponding batch size is bounded by the subsample size and the
        # memory size.
        max_bs = min(max(n_subsamples / 5, mG), n_subsamples)
        n_components = np.sum(np.power(1 / E, alpha) < max_bs) - 1
        if n_components < 2:
            n_components = min(E.shape[0] - 1, 2)

        Lambda = Lambda[:, :n_components]
        scale = np.power(E[0] / E[n_components], alpha)

        # Compute part of the preconditioner for step 2 of gradient descent in
        # the eigenpro model
        D = (1 - np.power(E[n_components] / E[:n_components], alpha)) / E[
            :n_components
        ]

        max_S = E[0].astype(np.float32)
        kxx = 1 - np.sum(Lambda**2, axis=1) * n_subsamples
        return max_S / scale, np.max(kxx), D, Lambda

    def _initialize_params(self, X, Y, random_state):
        """
        Validate parameters passed to the model, choose parameters
        that have not been passed in, and run setup for EigenPro iteration.
        Parameters
        ----------
        X : {float, array}, shape = [n_samples, n_features]
            Training data.

        Y : {float, array}, shape = [n_samples, n_targets]
            Training targets.

        random_state : RandomState instance
            The random state to use for random number generation

        Returns
        -------
        Y : {float, array}, shape = [n_samples, n_targets]
            Training targets. If Y was originally of shape
            [n_samples], it is now [n_samples, 1].

        E : {float, array}, shape = [k]
            Preconditioner for EigenPro

        Lambda : {float, array}, shape = [n_subsamples, k]
            Top eigenvectors of a subsample kernel matrix

        eta : float
            The learning rate

        pinx : {int, array}, shape = [sample_size]
            The rows of X used to calculate E and Lambda
        """
        n, d = X.shape
        n_label = 1 if len(Y.shape) == 1 else Y.shape[1]
        self.centers_ = X

        # Calculate the subsample size to be used.
        if self.subsample_size == "auto":
            if n < 100000:
                sample_size = 4000
            else:
                sample_size = 12000
        else:
            sample_size = self.subsample_size
        sample_size = min(n, sample_size)

        n_components = min(sample_size - 1, self.n_components)
        n_components = max(1, n_components)

        # Approximate amount of memory that we want to use
        mem_bytes = 0.1 * 1024**3
        # Memory used with a certain sample size
        mem_usages = (d + n_label + 2 * np.arange(sample_size)) * n * 4
        mG = np.int32(np.sum(mem_usages < mem_bytes))

        # Calculate largest eigenvalue and max{k(x,x)} using subsamples.
        pinx = random_state.choice(n, sample_size, replace=False).astype(
            "int32"
        )
        if self.gamma == "scale":
            self.gamma_ = np.float32(1.0 / (X.var() * d))
        else:
            self.gamma_ = self.gamma
        max_S, beta, E, Lambda = self._setup(
            X[pinx], n_components, mG, alpha=0.95
        )
        # Calculate best batch size.
        if self.batch_size == "auto":
            bs = min(np.int32(beta / max_S), mG) + 1
        else:
            bs = self.batch_size
        self.bs_ = min(bs, n)

        # Calculate best step size.
        if self.bs_ < beta / max_S + 1:
            eta = self.bs_ / beta
        elif self.bs_ < n:
            eta = 2.0 * self.bs_ / (beta + (self.bs_ - 1) * max_S)
        else:
            eta = 0.95 * 2 / max_S
        # Remember the shape of Y for predict() and ensure it's shape is 2-D.
        self.was_1D_ = False
        if len(Y.shape) == 1:
            Y = np.reshape(Y, (Y.shape[0], 1))
            self.was_1D_ = True
        return Y, E, Lambda, np.float32(eta), pinx

    def validate_parameters(self):
        """
        Validate the parameters of the model to ensure that no unreasonable
        values were passed in.
        """
        if self.n_epoch <= 0:
            raise ValueError(
                "n_epoch should be positive, was " + str(self.n_epoch)
            )
        if self.n_components < 0:
            raise ValueError(
                "n_components should be non-negative, was "
                + str(self.n_components)
            )
        if self.subsample_size != "auto" and self.subsample_size < 0:
            raise ValueError(
                "subsample_size should be non-negative, was "
                + str(self.subsample_size)
            )
        if self.batch_size != "auto" and self.batch_size <= 0:
            raise ValueError(
                "batch_size should be positive, was " + str(self.batch_size)
            )
        if self.gamma != "scale" and self.gamma <= 0:
            raise ValueError(
                "gamma should be positive, was " + str(self.gamma)
            )

    def _raw_fit(self, X, Y):
        """Train eigenpro regression model

        Parameters
        ----------
        X : {float, array}, shape = [n_samples, n_features]
            Training data.

        Y : {float, array}, shape = [n_samples, n_targets]
            Training targets.

        Returns
        -------
        self : returns an instance of self.
        """
        X, Y = check_X_y(
            X,
            Y,
            dtype=np.float32,
            multi_output=True,
            ensure_min_samples=3,
            y_numeric=True,
        )
        Y = Y.astype(np.float32)
        random_state = check_random_state(self.random_state)

        self.validate_parameters()
        """Parameter Initialization"""
        Y, D, V, eta, pinx = self._initialize_params(X, Y, random_state)

        """Training loop"""
        n = self.centers_.shape[0]

        self.coef_ = np.zeros((n, Y.shape[1]), dtype=np.float32)
        step = np.float32(eta / self.bs_)
        for _ in range(0, self.n_epoch):
            epoch_inds = random_state.choice(
                n, n // self.bs_ * self.bs_, replace=False
            ).astype("int32")

            for batch_inds in np.array_split(epoch_inds, n // self.bs_):
                batch_x = self.centers_[batch_inds]
                kfeat = self._kernel(batch_x, self.centers_)
                batch_y = Y[batch_inds]

                # Update 1: Sampled Coordinate Block.
                gradient = np.dot(kfeat, self.coef_) - batch_y

                self.coef_[batch_inds] -= step * gradient

                # Update 2: Fixed Coordinate Block
                delta = np.dot(
                    V * D, np.dot(V.T, np.dot(kfeat[:, pinx].T, gradient))
                )
                self.coef_[pinx] += step * delta
        return self

    def _raw_predict(self, X):
        """Predict using the kernel regression model

        Parameters
        ----------
        X : {float, array}, shape = [n_samples, n_features]
            Samples.

        Returns
        -------
        Y : {float, array}, shape = [n_samples, n_targets]
            Predicted targets.
        """
        check_is_fitted(
            self, ["bs_", "centers_", "coef_", "was_1D_", "gamma_"]
        )
        X = np.asarray(X, dtype=np.float64)

        if len(X.shape) == 1:
            raise ValueError(
                "Reshape your data. X should be a matrix of shape"
                " (n_samples, n_features)."
            )
        n = X.shape[0]

        Ys = []
        for batch_inds in np.array_split(range(n), max(1, n // self.bs_)):
            batch_x = X[batch_inds]
            kfeat = self._kernel(batch_x, self.centers_)

            pred = np.dot(kfeat, self.coef_)
            Ys.append(pred)
        Y = np.vstack(Ys)
        if self.was_1D_:
            Y = np.reshape(Y, Y.shape[0])
        return Y

    def _get_tags(self):
        tags = super()._get_tags()
        tags["multioutput"] = True
        return tags


class EigenProRegressor(RegressorMixin, BaseEigenPro):
    """Regression using EigenPro iteration.

    Train least squared kernel regression model with mini-batch EigenPro
    iteration.

    Parameters
    ----------
    batch_size : int, default = 'auto'
        Mini-batch size for gradient descent.

    n_epoch : int, default = 2
        The number of passes over the training data.

    n_components : int, default = 1000
        the maximum number of eigendirections used in modifying the kernel
        operator. Convergence rate speedup over normal gradient descent is
        approximately the largest eigenvalue over the n_componentth
        eigenvalue, however, it may take time to compute eigenvalues for
        large n_components

    subsample_size : int, default = 'auto'
        The number of subsamples used for estimating the largest
        n_component eigenvalues and eigenvectors. When it is set to 'auto',
        it will be 4000 if there are less than 100,000 samples
        (for training), and otherwise 12000.

    kernel : string or callable, default = "rbf"
        Kernel mapping used internally. Strings can be anything supported
        by scikit-learn, however, there is special support for the
        rbf, laplace, and cauchy kernels. If a callable is given, it should
        accept two arguments and return a floating point number.

    gamma : float, default='scale'
        Kernel coefficient. If 'scale', gamma = 1/(n_features*X.var()).
        Interpretation of the default value is left to the kernel;
        see the documentation for sklearn.metrics.pairwise.
        For kernels that use bandwidth, bandwidth = 1/sqrt(2*gamma).

    degree : float, default=3
        Degree of the polynomial kernel. Ignored by other kernels.

    coef0 : float, default=1
        Zero coefficient for polynomial and sigmoid kernels.
        Ignored by other kernels.

    kernel_params : mapping of string to any
        Additional parameters (keyword arguments) for kernel function
        passed as callable object.

    random_state : int, RandomState instance or None, (default=None)
        The seed of the pseudo random number generator to use when
        shuffling the data.  If int, random_state is the seed used by the
        random number generator; If RandomState instance, random_state is
        the random number generator; If None, the random number generator
        is the RandomState instance used by `np.random`.

    References
    ----------
    * Siyuan Ma, Mikhail Belkin
      "Diving into the shallows: a computational perspective on
      large-scale machine learning", NIPS 2017.

    Examples
    --------
    >>> from sklearn_extra.kernel_methods import EigenProRegressor
    >>> import numpy as np
    >>> n_samples, n_features, n_targets = 4000, 20, 3
    >>> rng = np.random.RandomState(1)
    >>> x_train = rng.randn(n_samples, n_features)
    >>> y_train = rng.randn(n_samples, n_targets)
    >>> rgs = EigenProRegressor(n_epoch=3, gamma=.5, subsample_size=50)
    >>> rgs.fit(x_train, y_train)
    EigenProRegressor(gamma=0.5, n_epoch=3, subsample_size=50)
    >>> y_pred = rgs.predict(x_train)
    >>> loss = np.mean(np.square(y_train - y_pred))
    """

    def __init__(
        self,
        batch_size="auto",
        n_epoch=2,
        n_components=1000,
        subsample_size="auto",
        kernel="rbf",
        gamma="scale",
        degree=3,
        coef0=1,
        kernel_params=None,
        random_state=None,
    ):
        super().__init__(
            batch_size=batch_size,
            n_epoch=n_epoch,
            n_components=n_components,
            subsample_size=subsample_size,
            kernel=kernel,
            gamma=gamma,
            degree=degree,
            coef0=coef0,
            kernel_params=kernel_params,
            random_state=random_state,
        )

    def fit(self, X, Y):
        return self._raw_fit(X, Y)

    def predict(self, X):
        return self._raw_predict(X)


class EigenProClassifier(ClassifierMixin, BaseEigenPro):
    """Classification using EigenPro iteration.

    Train least squared kernel classification model with mini-batch EigenPro
    iteration.

    Parameters
    ----------
    batch_size : int, default = 'auto'
        Mini-batch size for gradient descent.

    n_epoch : int, default = 2
        The number of passes over the training data.

    n_components : int, default = 1000
        the maximum number of eigendirections used in modifying the
        kernel operator. Convergence rate speedup over normal gradient
        descent is approximately the largest eigenvalue over the
        n_componenth eigenvalue, however, it may take time to compute
        eigenvalues for large n_components

    subsample_size : int, default = 'auto'
        The size of subsamples used for estimating the largest
        n_component eigenvalues and eigenvectors. When it is set to
        'auto', it will be 4000 if there are less than 100,000 samples
        (for training), and otherwise 12000.

    kernel : string or callable, default = "rbf"
        Kernel mapping used internally. Strings can be anything supported
        by scikit-learn, however, there is special support for the
        rbf, laplace, and cauchy kernels. If a callable is given, it should
        accept two arguments and return a floating point number.

    gamma : float, default='scale'
        Kernel coefficient. If 'scale', gamma = 1/(n_features*X.var()).
        Interpretation of the default value is left to the kernel;
        see the documentation for sklearn.metrics.pairwise.
        For kernels that use bandwidth, bandwidth = 1/sqrt(2*gamma).

    degree : float, default=3
        Degree of the polynomial kernel. Ignored by other kernels.

    coef0 : float, default=1
        Zero coefficient for polynomial and sigmoid kernels. Ignored by
        other kernels.

    kernel_params : mapping of string to any
        Additional parameters (keyword arguments) for kernel function
        passed as callable object.

    random_state : int, RandomState instance or None (default=None)
        The seed of the pseudo random number generator to use when
        shuffling the data.  If int, random_state is the seed used by
        the random number generator; If RandomState instance,
        random_state is the random number generator;
        If None, the random number generator is the RandomState
        instance used by `np.random`.

    References
    ----------
    * Siyuan Ma, Mikhail Belkin
      "Diving into the shallows: a computational perspective on
      large-scale machine learning", NIPS 2017.

    Examples
    --------
    >>> from sklearn_extra.kernel_methods import EigenProClassifier
    >>> import numpy as np
    >>> n_samples, n_features, n_targets = 4000, 20, 3
    >>> rng = np.random.RandomState(1)
    >>> x_train = rng.randn(n_samples, n_features)
    >>> y_train = rng.randint(n_targets, size=n_samples)
    >>> rgs = EigenProClassifier(n_epoch=3, gamma=.01, subsample_size=50)
    >>> rgs.fit(x_train, y_train)
    EigenProClassifier(gamma=0.01, n_epoch=3, subsample_size=50)
    >>> y_pred = rgs.predict(x_train)
    >>> loss = np.mean(y_train != y_pred)
    """

    def __init__(
        self,
        batch_size="auto",
        n_epoch=2,
        n_components=1000,
        subsample_size="auto",
        kernel="rbf",
        gamma=0.02,
        degree=3,
        coef0=1,
        kernel_params=None,
        random_state=None,
    ):
        super().__init__(
            batch_size=batch_size,
            n_epoch=n_epoch,
            n_components=n_components,
            subsample_size=subsample_size,
            kernel=kernel,
            gamma=gamma,
            degree=degree,
            coef0=coef0,
            kernel_params=kernel_params,
            random_state=random_state,
        )

    def fit(self, X, Y):
        """Train eigenpro classification model

        Parameters
        ----------
        X : {float, array}, shape = [n_samples, n_raw_feature]
            The raw input feature matrix.

        Y : {float, array}, shape =[n_samples]
            The labels corresponding to the features of X.

        Returns
        -------
        self : returns an instance of self.
        """
        X, Y = check_X_y(
            X,
            Y,
            dtype=np.float32,
            force_all_finite=True,
            multi_output=False,
            ensure_min_samples=3,
        )
        check_classification_targets(Y)
        self.classes_ = np.unique(Y)

        loc = {}
        for ind, label in enumerate(self.classes_):
            loc[label] = ind

        class_matrix = np.zeros((Y.shape[0], self.classes_.shape[0]))

        for ind, label in enumerate(Y):
            class_matrix[ind, loc[label]] = 1
        self._raw_fit(X, class_matrix)
        return self

    def predict(self, X):
        """Predict using the kernel classification model

        Parameters
        ----------
        X : {float, array}, shape = [n_samples, n_features]
            Samples.

        Returns
        -------
        y : {float, array}, shape = [n_samples]
            Predicted labels.
        """
        Y = self._raw_predict(X)
        return self.classes_[np.argmax(Y, axis=1)]



"""Tests for common-nearest neighbour clustering
"""

import pickle

import numpy as np
from numpy.testing import assert_array_equal

# TODO: Use
# from sklearn.utils._testing import assert_array_equal
#     not in scikit-learn version 0.21.3

from scipy.spatial import distance
from scipy import sparse

import pytest

from sklearn.neighbors import NearestNeighbors
from sklearn_extra.cluster import CommonNNClustering
from sklearn_extra.cluster import commonnn
from sklearn.cluster.tests.common import generate_clustered_data
from sklearn.metrics.pairwise import pairwise_distances


# TODO Tests where adapted from sklearn.cluster.tests.test_dbscan
#     of scikit-learn version 0.24.dev0.
#     To make sklearn_extra.cluster._commonnn compatible with
#     scikit-learn version 0.21.1 changes have been made
#     (see sklearn_extra.cluster._commonnn), e.g. regarding the input
#     validation.  Tests failing after the changes when calculating
#     neighbourhoods are skipped for now
#     with reason INPUT_VALIDATION_REASON
INPUT_VALIDATION_REASON = (
    "Broken after change of input validation with check_array"
)

n_clusters = 3
X = generate_clustered_data(n_clusters=n_clusters)


def test_commonnn_similarity():
    # Tests the algorithm with a similarity array.
    # Parameters chosen specifically for this task.
    eps = 0.15
    min_samples = 5
    # Compute similarities
    D = distance.squareform(distance.pdist(X))
    D /= np.max(D)
    # Compute
    labels = commonnn(
        D, metric="precomputed", eps=eps, min_samples=min_samples
    )
    # number of clusters, ignoring noise if present
    n_clusters_1 = len(set(labels)) - (1 if -1 in labels else 0)

    assert n_clusters_1 == n_clusters

    cobj = CommonNNClustering(
        metric="precomputed", eps=eps, min_samples=min_samples
    )
    labels = cobj.fit(D).labels_

    n_clusters_2 = len(set(labels)) - int(-1 in labels)
    assert n_clusters_2 == n_clusters


def test_cnn_feature():
    # Tests the algorithm with a feature vector array.
    # Parameters chosen specifically for this task.
    # Different eps to other test, because distance is not normalised.
    eps = 0.8
    min_samples = 5
    metric = "euclidean"
    # Compute
    # parameters chosen for task
    labels = commonnn(X, metric=metric, eps=eps, min_samples=min_samples)

    # number of clusters, ignoring noise if present
    n_clusters_1 = len(set(labels)) - int(-1 in labels)
    assert n_clusters_1 == n_clusters

    cobj = CommonNNClustering(metric=metric, eps=eps, min_samples=min_samples)
    labels = cobj.fit(X).labels_

    n_clusters_2 = len(set(labels)) - int(-1 in labels)
    assert n_clusters_2 == n_clusters


def test_cnn_sparse():
    labels_sparse = commonnn(sparse.lil_matrix(X), eps=0.8, min_samples=5)
    labels_dense = commonnn(X, eps=0.8, min_samples=5)
    assert_array_equal(labels_dense, labels_sparse)


@pytest.mark.skip(reason=INPUT_VALIDATION_REASON)
@pytest.mark.parametrize("include_self", [False, True])
def test_cnn_sparse_precomputed(include_self):
    D = pairwise_distances(X)
    nn = NearestNeighbors(radius=0.9).fit(X)
    X_ = X if include_self else None
    D_sparse = nn.radius_neighbors_graph(X=X_, mode="distance")
    # Ensure it is sparse not merely on diagonals:
    assert D_sparse.nnz < D.shape[0] * (D.shape[0] - 1)
    labels_sparse = commonnn(
        D_sparse, eps=0.8, min_samples=5, metric="precomputed"
    )
    labels_dense = commonnn(D, eps=0.8, min_samples=5, metric="precomputed")
    assert_array_equal(labels_dense, labels_sparse)


@pytest.mark.skip(reason=INPUT_VALIDATION_REASON)
def test_cnn_sparse_precomputed_different_eps():
    # test that precomputed neighbors graph is filtered if computed with
    # a radius larger than eps.
    lower_eps = 0.2
    nn = NearestNeighbors(radius=lower_eps).fit(X)
    D_sparse = nn.radius_neighbors_graph(X, mode="distance")
    cnn_lower = commonnn(D_sparse, eps=lower_eps, metric="precomputed")

    higher_eps = lower_eps + 0.7
    nn = NearestNeighbors(radius=higher_eps).fit(X)
    D_sparse = nn.radius_neighbors_graph(X, mode="distance")
    cnn_higher = commonnn(D_sparse, eps=lower_eps, metric="precomputed")

    assert_array_equal(cnn_lower, cnn_higher)


@pytest.mark.skip(reason=INPUT_VALIDATION_REASON)
@pytest.mark.parametrize("use_sparse", [True, False])
@pytest.mark.parametrize("metric", ["precomputed", "minkowski"])
def test_cnn_input_not_modified(use_sparse, metric):
    # test that the input is not modified by cnn
    X = np.random.RandomState(0).rand(10, 10)
    X = sparse.csr_matrix(X) if use_sparse else X
    X_copy = X.copy()
    commonnn(X, metric=metric)

    if use_sparse:
        assert_array_equal(X.toarray(), X_copy.toarray())
    else:
        assert_array_equal(X, X_copy)


def test_cnn_callable():
    # Tests the algorithm with a callable metric.
    # Parameters chosen specifically for this task.
    # Different eps to other test, because distance is not normalised.
    eps = 0.8
    min_samples = 5
    # metric is the function reference, not the string key.
    metric = distance.euclidean
    # Compute
    # parameters chosen for task
    labels = commonnn(
        X,
        metric=metric,
        eps=eps,
        min_samples=min_samples,
        algorithm="ball_tree",
    )

    # number of clusters, ignoring noise if present
    n_clusters_1 = len(set(labels)) - int(-1 in labels)
    assert n_clusters_1 == n_clusters

    cobj = CommonNNClustering(
        metric=metric, eps=eps, min_samples=min_samples, algorithm="ball_tree"
    )
    labels = cobj.fit(X).labels_

    n_clusters_2 = len(set(labels)) - int(-1 in labels)
    assert n_clusters_2 == n_clusters


def test_cnn_metric_params():
    # Tests that clustering works with the metrics_params argument.
    eps = 0.8
    min_samples = 5
    p = 1

    # Compute DBSCAN with metric_params arg
    cobj = CommonNNClustering(
        metric="minkowski",
        metric_params={"p": p},
        eps=eps,
        min_samples=min_samples,
        algorithm="ball_tree",
    ).fit(X)
    labels_1 = cobj.labels_

    # Test that sample labels are the same as passing Minkowski 'p' directly
    cobj = CommonNNClustering(
        metric="minkowski",
        eps=eps,
        min_samples=min_samples,
        algorithm="ball_tree",
        p=p,
    ).fit(X)
    labels_2 = cobj.labels_

    assert_array_equal(labels_1, labels_2)

    # Minkowski with p=1 should be equivalent to Manhattan distance
    cobj = CommonNNClustering(
        metric="manhattan",
        eps=eps,
        min_samples=min_samples,
        algorithm="ball_tree",
    ).fit(X)
    labels_3 = cobj.labels_

    assert_array_equal(labels_1, labels_3)


def test_cnn_balltree():
    # Tests the algorithm with balltree for neighbor calculation.
    eps = 0.8
    min_samples = 5

    D = pairwise_distances(X)
    labels = commonnn(
        D, metric="precomputed", eps=eps, min_samples=min_samples
    )

    # number of clusters, ignoring noise if present
    n_clusters_1 = len(set(labels)) - int(-1 in labels)
    assert n_clusters_1 == n_clusters

    cobj = CommonNNClustering(
        p=2.0, eps=eps, min_samples=min_samples, algorithm="ball_tree"
    )
    labels = cobj.fit(X).labels_

    n_clusters_2 = len(set(labels)) - int(-1 in labels)
    assert n_clusters_2 == n_clusters

    cobj = CommonNNClustering(
        p=2.0, eps=eps, min_samples=min_samples, algorithm="kd_tree"
    )
    labels = cobj.fit(X).labels_

    n_clusters_3 = len(set(labels)) - int(-1 in labels)
    assert n_clusters_3 == n_clusters

    cobj = CommonNNClustering(
        p=1.0, eps=eps, min_samples=min_samples, algorithm="ball_tree"
    )
    labels = cobj.fit(X).labels_

    n_clusters_4 = len(set(labels)) - int(-1 in labels)
    assert n_clusters_4 == n_clusters

    cobj = CommonNNClustering(
        leaf_size=20, eps=eps, min_samples=min_samples, algorithm="ball_tree"
    )
    labels = cobj.fit(X).labels_

    n_clusters_5 = len(set(labels)) - int(-1 in labels)
    assert n_clusters_5 == n_clusters


def test_input_validation():
    # CommonNNClustering.fit should accept a list of lists.
    X = [[1.0, 2.0], [3.0, 4.0]]
    CommonNNClustering().fit(X)  # must not raise exception


@pytest.mark.parametrize(
    "args",
    [
        {"eps": -1.0},
        {"algorithm": "blah"},
        {"metric": "blah"},
        {"leaf_size": -1},
        {"p": -1},
    ],
)
def test_cnn_badargs(args):
    # Test bad argument values: these should all raise ValueErrors
    with pytest.raises(ValueError):
        commonnn(X, **args)


def test_pickle():
    obj = CommonNNClustering()
    s = pickle.dumps(obj)
    assert type(pickle.loads(s)) == obj.__class__


def test_boundaries():
    # ensure min_samples is inclusive of core point
    core = np.where(commonnn([[0], [1]], eps=2, min_samples=0) >= 0)[0]
    assert 0 in core
    # ensure eps is inclusive of circumference
    core = np.where(commonnn([[0], [1], [1]], eps=1, min_samples=0) >= 0)[0]
    assert 0 in core
    core = np.where(commonnn([[0], [1], [1]], eps=0.99, min_samples=0) >= 0)[0]
    assert 0 not in core


@pytest.mark.skip(reason="Sample weights have no effect on the clustering yet")
def test_weighted_cnn():
    # ensure sample_weight is validated
    with pytest.raises(ValueError):
        commonnn([[0], [1]], sample_weight=[2])
    with pytest.raises(ValueError):
        commonnn([[0], [1]], sample_weight=[2, 3, 4])

    # ensure sample_weight has an effect
    assert_array_equal(
        [], commonnn([[0], [1]], sample_weight=None, min_samples=6)[0]
    )
    assert_array_equal(
        [], commonnn([[0], [1]], sample_weight=[5, 5], min_samples=6)[0]
    )
    assert_array_equal(
        [0], commonnn([[0], [1]], sample_weight=[6, 5], min_samples=6)[0]
    )
    assert_array_equal(
        [0, 1], commonnn([[0], [1]], sample_weight=[6, 6], min_samples=6)[0]
    )

    # points within eps of each other:
    assert_array_equal(
        [0, 1],
        commonnn([[0], [1]], eps=1.5, sample_weight=[5, 1], min_samples=6)[0],
    )
    # and effect of non-positive and non-integer sample_weight:
    assert_array_equal(
        [],
        commonnn([[0], [1]], sample_weight=[5, 0], eps=1.5, min_samples=6)[0],
    )
    assert_array_equal(
        [0, 1],
        commonnn([[0], [1]], sample_weight=[5.9, 0.1], eps=1.5, min_samples=6)[
            0
        ],
    )
    assert_array_equal(
        [0, 1],
        commonnn([[0], [1]], sample_weight=[6, 0], eps=1.5, min_samples=6)[0],
    )
    assert_array_equal(
        [],
        commonnn([[0], [1]], sample_weight=[6, -1], eps=1.5, min_samples=6)[0],
    )

    # for non-negative sample_weight, cores should be identical to repetition
    rng = np.random.RandomState(42)
    sample_weight = rng.randint(0, 5, X.shape[0])
    label1 = commonnn(X, sample_weight=sample_weight)
    assert len(label1) == len(X)

    X_repeated = np.repeat(X, sample_weight, axis=0)
    label_repeated = commonnn(X_repeated)
    core_repeated_mask = np.zeros(X_repeated.shape[0], dtype=bool)
    core_repeated_mask[np.where(label_repeated >= 0)[0]] = True
    core_mask = np.zeros(X.shape[0], dtype=bool)
    core_mask[np.where(label1 >= 0)[0]] = True
    assert_array_equal(np.repeat(core_mask, sample_weight), core_repeated_mask)

    # sample_weight should work with precomputed distance matrix
    D = pairwise_distances(X)
    core3, label3 = commonnn(
        D, sample_weight=sample_weight, metric="precomputed"
    )
    assert_array_equal(label1, label3)

    # sample_weight should work with estimator
    est = CommonNNClustering().fit(X, sample_weight=sample_weight)
    label4 = est.labels_
    assert_array_equal(label1, label4)

    est = CommonNNClustering()
    label5 = est.fit_predict(X, sample_weight=sample_weight)

    assert_array_equal(label1, label5)
    assert_array_equal(label1, est.labels_)


@pytest.mark.parametrize("algorithm", ["brute", "kd_tree", "ball_tree"])
def test_cnn_core_samples_toy_1(algorithm):
    X = [[0], [2], [3], [4], [6], [8], [10]]

    # Within eps = 1, only points at 2, 3, and 4
    # are neighbours. Valid clusters need to have more than one
    # members, so all other points are isolated and considered
    # noise.
    labels = commonnn(X, algorithm=algorithm, eps=1, min_samples=0)
    assert_array_equal(labels, [-1, 0, 0, 0, -1, -1, -1])

    # With eps=1 and min_samples=1 the 3 samples from the
    # denser area are no core samples anymore (2 and 4 share 3 as
    # common neighbour but are not neighbours of each other)
    labels = commonnn(X, algorithm=algorithm, eps=1, min_samples=1)
    assert_array_equal(labels, [-1, -1, -1, -1, -1, -1, -1])


@pytest.mark.parametrize("algorithm", ["brute", "kd_tree", "ball_tree"])
def test_cnn_core_samples_toy_2(algorithm):
    X = [
        [0, 0],  # 0
        [1, 1],  # 1
        [1, 0],  # 2
        [0, -1],  # 3
        [0.5, -0.5],  # 4
        [2, 1.5],  # 5
        [2.5, -0.5],  # 6
        [4, 2],  # 7
        [4.5, 2.5],  # 8
        [5, -1],  # 9
        [5.5, -0.5],  # 10
        [5.5, -1.5],
    ]  # 11

    labels = commonnn(X, algorithm=algorithm, eps=1.5, min_samples=0)
    assert_array_equal(labels, [0, 0, 0, 0, 0, 0, -1, 1, 1, 2, 2, 2])

    labels = commonnn(X, algorithm=algorithm, eps=1.5, min_samples=1)
    assert_array_equal(labels, [0, 0, 0, 0, 0, -1, -1, -1, -1, 1, 1, 1])

    labels = commonnn(X, algorithm=algorithm, eps=1.5, min_samples=2)
    assert_array_equal(labels, [0, -1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1])

    labels = commonnn(X, algorithm=algorithm, eps=1.5, min_samples=3)
    assert_array_equal(labels, [0, -1, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1])

    labels = commonnn(X, algorithm=algorithm, eps=1.5, min_samples=4)
    assert_array_equal(
        labels, [-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1]
    )


def test_cnn_precomputed_metric_with_degenerate_input_arrays():
    # see https://github.com/scikit-learn/scikit-learn/issues/4641 for
    # more details
    X = np.eye(10)
    labels = CommonNNClustering(eps=0.5, metric="precomputed").fit(X).labels_
    assert len(set(labels)) == 1

    X = np.zeros((10, 10))
    labels = CommonNNClustering(eps=0.5, metric="precomputed").fit(X).labels_
    assert len(set(labels)) == 1


@pytest.mark.skip(reason=INPUT_VALIDATION_REASON)
def test_cnn_precomputed_metric_with_initial_rows_zero():
    # sample matrix with initial two row all zero
    ar = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0],
            [0.0, 0.0, 0.1, 0.1, 0.0, 0.0, 0.3],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1],
            [0.0, 0.0, 0.0, 0.0, 0.3, 0.1, 0.0],
        ]
    )
    matrix = sparse.csr_matrix(ar)
    labels = (
        CommonNNClustering(eps=0.2, metric="precomputed", min_samples=0)
        .fit(matrix)
        .labels_
    )
    assert_array_equal(labels, [-1, -1, 0, 0, 0, 1, 1])



import datetime

import numpy as np

from sklearn.kernel_approximation import Fastfood
from sklearn.kernel_approximation import RBFSampler
from sklearn.utils.testing import assert_greater

"""compares the performance of Fastfood and RKS"""
# generate data
rng = np.random.RandomState(0)
X = rng.random_sample(size=(1000, 4096))
Y = rng.random_sample(size=(10000, 4096))
X /= X.sum(axis=1)[:, np.newaxis]
Y /= Y.sum(axis=1)[:, np.newaxis]

# calculate feature maps
gamma = 10.0
sigma = np.sqrt(1 / (2 * gamma))
number_of_features_to_generate = 4096 * 4

exact_start = datetime.datetime.utcnow()
# original rbf kernel method:
# rbf_kernel(X, X, gamma=gamma)
# rbf_kernel(X, Y, gamma=gamma)
exact_end = datetime.datetime.utcnow()
exact_spent_time = exact_end - exact_start
print("Timimg exact rbf: \t\t", exact_spent_time)

rbf_transform = Fastfood(
    sigma=sigma,
    n_components=number_of_features_to_generate,
    tradeoff_mem_accuracy="mem",
    random_state=42,
)
_ = rbf_transform.fit(X)
fastfood_fast_vec_start = datetime.datetime.utcnow()
# Fastfood: approximate kernel mapping
_ = rbf_transform.transform(X)
_ = rbf_transform.transform(Y)
fastfood_fast_vec_end = datetime.datetime.utcnow()
fastfood_fast_vec_spent_time = fastfood_fast_vec_end - fastfood_fast_vec_start
print("Timimg fastfood fast vectorized: \t\t", fastfood_fast_vec_spent_time)

rks_rbf_transform = RBFSampler(
    gamma=gamma, n_components=number_of_features_to_generate, random_state=42
)
_ = rks_rbf_transform.fit(X)
rks_start = datetime.datetime.utcnow()
# Random Kitchens Sinks: approximate kernel mapping
_ = rks_rbf_transform.transform(X)
_ = rks_rbf_transform.transform(Y)
rks_end = datetime.datetime.utcnow()
rks_spent_time = rks_end - rks_start
print("Timimg rks: \t\t\t", rks_spent_time)

assert_greater(rks_spent_time, fastfood_fast_vec_spent_time)
